/* crypt_rsa.h - NextIRCd
 * Copyright (C) 2012-2013  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef _crypt_rsa_h
#define _crypt_rsa_h

struct crypt_rsa_pubkey;
struct crypt_rsa_privkey;

void crypt_init();
void crypt_deinit();

struct crypt_rsa_pubkey *crypt_rsa_load_pubkey(const char *key);
void crypt_rsa_unload_pubkey(struct crypt_rsa_pubkey *pubkey);
char *crypt_rsa_export_pubkey(struct crypt_rsa_pubkey *privkey);
struct crypt_rsa_privkey *crypt_rsa_load_privkey(const char *key);
struct crypt_rsa_privkey *crypt_rsa_generate_privkey(const int keysize);
void crypt_rsa_unload_privkey(struct crypt_rsa_privkey *privkey);
char *crypt_rsa_export_privkey(struct crypt_rsa_privkey *privkey, int pubkey);
struct crypt_rsa_pubkey *crypt_rsa_get_pubkey(struct crypt_rsa_privkey *privkey);

int crypt_rsa_encrypt_data(struct crypt_rsa_pubkey *pubkey, const char *data, int datalen, char **encrypted);
int crypt_rsa_decrypt_data(struct crypt_rsa_privkey *privkey, const char *encrypted, int enclen, char **data);
int crypt_rsa_encrypt_maxlen(struct crypt_rsa_pubkey *pubkey);
void crypt_rsa_encrypt_free(char *encrypted);
void crypt_rsa_decrypt_free(char *data);

#endif
